/***** BEGIN License Block *****
 * Version 1.1 MPL
 *
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License").  You may not use this 
 * file except in compliance with the License. You may obtain a 
 * copy of the License at http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed on an 
 * "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express  
 * or implied. See the License for the specific language governing 
 * rights and limitations under the License.
 *
 * The Original Code is the Civil Netizen/Civil Servant.
 *
 * The Initial Developer of the Original Code is Civil Engines 
 * Research. Portions created by Civil Engines Research LLC are 
 * Copyright (C) 2001-2003 Civil Engines Research.  All Rights 
 * Reserved.
 *
 * Redistribution of the Original Code or portions of the Original
 * Code or Software under this agreement must retain the above 
 * copyright notice, this redistribution statement and the following 
 * disclaimer.  Neither the name of Civil Engines Research nor the 
 * names of its contributors may be used to endorse or promote any
 * products derived from this software without specific prior 
 * written permission from Civil Engines Research.
 *
 *
 * Contributor(s): 
 *   Paolo de Dios <paolo.dedios@civilengines.com> (original author)
 *
 *
 ***** END License Block *****/


const NS_RDFSERIALIZER_CONTRACTID         = "@mozilla.org/rdf/xml-serializer;1";
const NS_RDFCONTAINER_CONTRACTID          = "@mozilla.org/rdf/container;1";


function ceRDFDataModelObserver()
{
}


ceRDFDataModelObserver.prototype =
{
    onBeginLoad : function(rdfResource)
    {
        Syslog.logMessage("Loading RDF resource");
    },

    onInterrupt : function(rdfResource)
    {
        Syslog.logMessage("RDF resource load interrupted");
    },
    
    onResume    : function(rdfResource)
    {
        Syslog.logMessage("Resuming RDF resource load");
    },
    
    onError     : function(rdfResource, status, errorMessage)
    {
        Syslog.logMessage("RDF resource load error => " + errorMessage);
    },

    onEndLoad   : function(rdfResource)
    {
        Syslog.logMessage("Finished loading RDF resource");
    }
};



function ceRDFXMLLoadObserver(rdfResource, rdfObserver)
{
    this.rdfResource = rdfResource;
    this.rdfObserver = rdfObserver;
}


ceRDFXMLLoadObserver.prototype =
{
    QueryInterface : function(iid)
    {
        if(iid.equals(Components.interfaces.nsIRDFXMLSinkObserver))
        {
            return this;
        }
        
        throw Components.results.NS_ERROR_NO_INTERFACE;
    },
    
    onBeginLoad : function(sink)
    {
        if(this.rdfObserver)
        {
            this.rdfObserver.onBeginLoad(this.rdfResource);
        }        
    },

    onInterrupt : function(sink)
    {
        if(this.rdfObserver)
        {
            this.rdfObserver.onInterrupt(this.rdfResource);
        }
    },
    
    onResume : function(sink)
    {
        if(this.rdfObserver)
        {
            this.rdfObserver.onResume(this.rdfResource);
        }
    },
    
    onError : function(sink, status, errorMessage)
    {
        if(this.rdfObserver)
        {
            this.rdfObserver.onError(this.rdfResource, status, errorMessage);
        }
    },

    onEndLoad : function(sink)
    {
        if(this.rdfObserver)
        {
            this.rdfObserver.onEndLoad(this.rdfResource);
        }
    }    
};


/* RDF is a directed label graph expresed as a series of 3-tuple 
 * statements consisting of 
 *
 *  [ statement ] --> [ property  ] --> [ target ]
 *
 * or in some circles also known as 
 *
 *  [ subject   ] --> [ predicate ] --> [ object ]
 *
 * Each resource, property and target are stored as a series of 
 * nodes in a data source following the RDF model. The abstract
 * RDF model implemented bellow is intended to abstract the 
 * API overhead of querying and invoking the RDF engine's XPCOM 
 * interfaces.
 *
 * The native XPCOM interfaces can be accessed via the following
 * properties:
 *
 *  nsIRDFDataSource                         :  ceRDFDataModel.dataSource
 *  nsIRDFDataSource -> nsIURI               :  ceRDFDataModel.uri
 *
 *  nsIRDFNode                               :  ceRDFNode.node 
 *    |                                         ceRDFResource.node
 *    |                                         ceRDFLiteral.node
 *    |
 *    +- nsIRDFResource                      :  ceRDFResource.node
 *    |
 *    +- nsIRDFResource -> nsIRDFDataSource  :  ceRDFResource.dataSource
 *    |
 *    |
 *    +- nsIRDFLiteral                       :  ceRDFLiteral.node
 *    |
 *    +- nsIRDFLiteral  -> nsIRDFDataSource  :  ceRDFLiteral.dataSource
 */
function ceRDFDataModel(uri, loadObserver)
{
    /* Create an in-memory data source if a datasource
     * URI is not provided by the caller.
     */
    if(uri)
    {
        this.uri = uri;
        this.loadDatasource(this.uri, loadObserver);
    }
    else
    {
        this.dataSource = do_CreateInstance(NS_RDFINMEMORYDATASOURCE_CONTRACTID   ,
                                            Components.interfaces.nsIRDFDataSource);
    }
}



ceRDFDataModel.prototype =
{
    /* uri        : null                      , */
    /* dataSource : null                      , */
    rdfService    : RDFService                ,
    rdfUtils      : RDFContainerUtils         ,
    rdfFactory    : new ceRDFResourceFactory(),

    getDatasource  : function()
    {
        return this.dataSource;
    },


    createResource : function(uri)
    {
        var resource        = rdfFactory.getResource(uri);
        resource.dataSource = this;
        
        return resource;
    },


    createAnonymousResource : function()
    {
        var resource        = rdfFactory.getAnonymousResource();
        resource.dataSource = this;

        return resource;
    },
    

    createLiteral : function(literalValue)
    {
        var literal         = rdfFactory.getLiteral(literalValue);
        literal.dataSource  = this;

        return resource;
    },


    addStatement : function(resource, property, target)
    {
        return this.dataSource.Assert(resource.node,
                                      property.node,
                                      target.node  ,
                                      true         );
    },


    removeStatement : function(resource, property, target)
    {
        return this.dataSource.Unassert(resource.node,
                                        property.node,
                                        target.node  );
    },


    containsStatement : function(resource, property, target)
    {
        if(resource.dataSource == this.dataSource &&
           property.dataSource == this.dataSource &&
           target.dataSource   == this.dataSource  )
        {
            return this.dataSource.HasAssertion(resource.node,
                                                property.node,
                                                target.node  ,
                                                true         );
        }
    },


    addStatementSet : function(statement)
    {
        return this.addStatement(statement.source  , 
                                 statement.property,
                                 statement.target  );
    },

    removeStatementSet : function(statement)
    {
        return this.removeStatement(statement.source  , 
                                    statement.property,
                                    statement.target  );
    },


    hasStatementSet : function(statement)
    {
        return this.containsStatement(statement.source  , 
                                      statement.property,
                                      statement.target  );
    },
    
    findStatementSource : function(property, target)
    {
        return target.getSource(property);
    },


    findAllStatementSources : function(property, target)
    {
        return target.getSources(property);
    },
    

    findStatementTarget : function(source, property)
    {
        return source.getTarget(property);
    },


    findAllStatementTargets : function(source, property)
    {
        return source.getTargets(property);
    },


    changeStatementTarget : function(source, property, oldTarget, newTarget)
    {
        this.dataSource.Change(source.node, property.node, oldTarget.node, newTarget.node);
    },
    

    changeStatementSource : function(oldSource, newSource, property, target)
    {
        this.dataSource.Move(oldSource.node, newSource.node, property.node, target.node);
    },


    getResourceEnumerator : function()
    {
        return new ceRDFNodeEnumerator(this.dataSource.GetAllResources(), this.dataSource);
    },
    
    
    getStatementCollectionWithSource : function(resource)
    {
        var property               = null;
        var target                 = null;
        var targetEnumeration      = null;
        var index                  = 0;
        var subjectStatementsArray = [];
        
        var propertyEnumeration    = resource.getPropertiesToTargets();
        while(propertyEnumeration.hasMoreElements())
        {
            property = targetEnumeration.getNext();
            
            targetEnumeration = resource.getTargets(property);
            while(targetEnumeration.hasMoreElements())
            {
                target = targetEnumeration.getNext();
                subjectStatementsArray[index++] = new ceRDFStatement(resource, predicate, target);
            }
        }
        
        return subjectStatementsArray;
    },
    

    getStatementCollection : function()
    {
        var statementArray = [];
        
        var subjectEnumeration = this.getResourceEnumerator();
        while(subjectEnumeration.hasMoreElements())
        {
            var subject             = subjectEnumeration.getNext();
            var subjectStatementSet = this.getStatementCollectionWithSource(subject);
            
            statementArray          = statementArray.concat(subjectStatementSet);
        }
                
        return statementArray;
    },
    

    createDataModelMatchingStatement : function(statement, modelURI)
    {
        var dataModel    = new ceRDFDataModel(modelURI);
        var statementSet = this.getStatementCollection();
        
        for(var i = 0, setlength = statementSet.length; i < setlength; i++)
        {
            var statementElement = statementSet[i];
            if(statementElement.equals(statement))
            {
                dataModel.addStatementSet(statementElement);
            }
        }
        
        return dataModel;
    },
    

    getDataModelRootedAt : function(resourceRoot, subModel)
    {
        if(resourceRoot instanceof ceRDFResource)
        {
            var target       = null;
            var statement    = new ceRDFStatement(resourceRoot, null, null);
            var modelMatch   = this.createModelMatchingStatement(statement);
            var statementSet = modelMatch.getStatementCollection()
             
            subBranch = this.rdfFactory.calculateModelUnion(subModel, modelMatch);
            for(var i = 0, setlength = statementSet.length; i < setlength; i++)
            {
                target = statementSet[i].target;
                if(target instanceof ceRDFResource)
                {
                    subModel = this.getDataModelRootedAt(target, subModel);
                }
            }
        }
        
        return subModel;
    },


    getResourceLabelsIn : function(rdfNode)
    {
        return rdfNode.getPropertiesFromSources();
    },


    getResourceLabelsOut : function(rdfNode)
    {
        if(rdfNode instanceof ceRDFLiteral)
        {
            return null;
        }
        
        return rdfNode.getPropertiesToTargets();
    },


    hasResourceLabelIn : function(rdfNode, property)
    {
        return rdfNode.hasPropertyFromSource(property);
    },


    hasResourceLabelOut : function(rdfNode, property)
    {
        if(rdfNode instanceof ceRDFLiteral)
        {
            return null;
        }
        
        return rdfNode.hasPropertyToTarget(property);
    },
    

    copyResourcesToDataModel : function(destinationModel)
    {
        if(this.dataSource == null)
        {
            this.dataSource = do_CreateInstance(NS_RDFINMEMORYDATASOURCE_CONTRACTID   ,
                                                Components.interfaces.nsIRDFDataSource);
        }
        

        if(destinationModel.dataSource == null)
        {
            destinationModel.dataSource = do_CreateInstance(NS_RDFINMEMORYDATASOURCE_CONTRACTID   ,
                                                            Components.interfaces.nsIRDFDataSource);
        }
        
        var srcDatasource  = this.dataSource;
        var destDatasource = destinationModel.dataSource;
        
        var sourceResourceEnumerator = srcDataSource.GetAllResources();
        
        while(sourceResourceEnumerator.hasMoreElements())
        {
            var source   = sourceResourceEnumerator.getNext();
            var property = srcDataSource.ArcLabelsOut(source);
            
            while(property.hasMoreElements())
            {
                var srcProperty = property.getNext();
                srcProperty = do_QueryInterface(srcProperty, Components.interfaces.nsIRDFResource);
                
                var target = srcDatasource.GetTarget(source, srcProperty, true);
                if(target != null)
                {
                    destDatasource.Assert(source, srcProperty, target, true);
                }
            }
        }
    },
    

    cloneDataModel : function()
    {
        var modelClone = new ceRDFDataModel();
        this.copyResourcesToDataModel(modelClone);
        
        return modelClone;
    },
    

    removeNativeDescendants : function(sourceNode, datasource)
    {
        var properties = datasource.ArcLabelsOut(sourceNode);
        while(properties.hasMoreElements())
        {
            var property = properties.getNext();
            var target   = nativeDataSource.GetTarget(sourceNode, property, true);
            
            try
            {
                target = do_QueryInterface(target, Components.interfaces.nsIRDFResource);
                this.removeNativeDescendatns(target, datasource);
            }
            catch(exception)
            {
            }
            
            datasource.Unassert(target, property, target);
        }        
    },

    
    removeResourceSubgraph : function(rdfResource)
    {
        if(this.dataSource == null)
        {
            return;
        }

        var node       = null;
        var datasource = this.dataSource;
                
        if(typeof(rdfResource) == "string")
        {
            node = this.rdfService.GetResource(rdfResource);
        }
        else
        {
            node = rdfResource.node;
        }
        
        this.removeNativeDescendatns(node, datasource);
        
        var properties = datasource.ArcLabelsIn(node);
        while(properties.hasMoreElements())
        {
            var property = properties.getNext();
            var source   = datasource.GetSource(property, node, true);
            
            datasource.Unassert(source, property, node);
        }        
    },    
   

    loadDatasource : function(uri, loadObserver)
    {
        if(uri.indexOf(":") == -1)
        {
            var documentURL = document.location.href;
            
            if(document.location.pathname == null)
            { 
                uri = documentURL + "/" + uri;
            }
            else
            { 
                uri = documentURL.substring(0, documentURL.lastIndexOf("/") + 1) + uri;
            }
        }
        
        if(loadObserver == null)
        {
            this.dataSource = this.rdfService.GetDataSourceBlocking(uri);
        }
        else 
        {
            this.dataSource       = this.rdfService.GetDataSource(uri);
            
            var remoteDataSource  = null;
            var dataSourceSink    = do_QueryInterface(this.dataSource                    ,
                                                      Components.interfaces.nsIRDFXMLSink);
            try 
            {
                remoteDataSource = do_QueryInterface(this.dataSource                             ,
                                                     Components.interfaces.nsIRDFRemoteDataSource);
            }
            catch(exception)
            {
                loadObserver.onEndLoad(this);
                return;
            }

            if(remoteDataSource.loaded)
            {
                loadObserver.onEndLoad(this);
            }
            
            var rdfXMLLoadObserver = new ceRDFXMLLoadObserver(this, loadObserver);
            dataSourceSink.addXMLSinkObserver(rdfXMLLoadObserver);
        }
    },


    flushToCString : function()
    {
        var stringOutputStream =
        {
            stringData    : ""   ,
            isNonBlocking : false,
            
            close : function() 
            {
            },
            
            flush : function() 
            {
            },
       
            write : function(stringBuffer, writeAmount) 
            {
                this.stringData += stringBuffer;
                return writeAmount;
            },

            writeFrom : function(outputStream, writeAmount) 
            {
            }
        };
    
        this.flushToStream(stringOutputStream);
        return stringOutputStream.stringData;
    },


    flushToStream : function(rdfOutputStream)
    {
        var rdfSerializer = do_CreateInstance(NS_RDFSERIALIZER_CONTRACTID              ,
                                              Components.interfaces.nsIRDFXMLSerializer);
        
        if(rdfSerializer)
        {
            rdfSerializer.init(this.dataSource);
            
            var rdfXMLSource = do_QueryInterface(rdfSerializer, Components.interfaces.nsIRDFXMLSource);
            if(rdfXMLSource)
            {
                rdfXMLSource.Serialize(rdfOutputStream);   
            }
        }
    },

    
    flushToURI : function(destinationURI)
    {
        var remoteDataSource = do_QueryInterface(this.dataSource                             ,
                                                 Components.interfaces.nsIRDFRemoteDataSource);
        if(remoteDatasource != null)
        {
            remoteDatasource.FlushTo(destinationURI);
            return true;
        }

        return false;
    },

    
    reload : function(synchronous)
    {
        var remoteDataSource = do_QueryInterface(this.datasource                             ,
                                                 Components.interfaces.nsIRDFRemoteDataSource);
        
        if(remoteDataSource != null)
        {
            remoteDataSource.Refresh(synchronous);
            return true;
        }
        
        return false;
    },
    

    flush : function()
    {
        var remoteDataSource = do_QueryInterface(this.datasource                             ,
                                                 Components.interfaces.nsIRDFRemoteDataSource);
        if(remoteDataSource != null)
        {
            remoteDataSource.Flush();
            return true;
        }
        
        return false;
    }
};




const RDFDataModelUtils =
{

    createMemoryDataModel : function(uri)
    {
        var memoryDataModel = new ceRDFDataModel();
        memoryDataModel.uri = uri;
        
        return memoryDataModel;
    },
    
    
    createSimpleLocalFileDataModel : function(filePath, namespaceNC, namespaceURI, loadObserver)
    {
        var fileDataModel = null;
        var filePathURI   = filePath;
        
        if(filePath.substr(0, 7) != "file://")
        {
            filePathURI = "file://" + filePath;
        }
        else
        {
            filePath    = filePath.substr(7);
        }
        
        var rdfFile = new File(filePath);
        
        if(!rdfFile.exists())
        {
            if(!rdfFile.open("w"))
            {
                return null;
            }
            
            var rdfFileInitialContent = 
                '<?xml version="1.0" ?>\n'                                        +
                '<rdf:RDF\n'                                                      +
                '     xmlns:' + namespaceNC + '="' + namespaceURI + '"\n'         +
                '     xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">\n' +
                '</rdf:RDF>\n';
            
            if(!rdfFile.write(rdfFileInitialContent))
            {
                rdfFile.close();
                return null;
            }
            
            rdfFile.close();
        }
        
        return new ceRDFDataModel(filePathURI, loadObserver);
    },
    
    
    /* namespaceSet = { ["nc1", "http://www.nc1.com/about#"] ,
     *                  ["nc2", "http://www.nc2.com/about#"] ,
     *                  ["nc3", "http://www.nc3.com/about#"] };
     */
    createLocalFileDataModel : function(filePath, namespaceSet, loadObserver)
    {
        var fileDataModel = null;
        var filePathURI   = filePath;
        
        if(filePath.substr(0, 7) != "file://")
        {
            filePathURI = "file://" + filePath;
        }
        else
        {
            filePath    = filePath.substr(7);
        }
        
        var rdfFile = new File(filePath);
        
        if(!rdfFile.exists())
        {
            if(!rdfFile.open("w"))
            {
                return null;
            }
            
            var rdfFileInitialContent = 
                '<?xml version="1.0" ?>\n'                                        +
                '<rdf:RDF\n'                                                      +
                '     xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">\n';
            
            for(var i = 0, nclength = namespaceSet.length; i < nclength; i++)
            {
                var namespacePair = namespaceSet[i];
                rdfFileInitialContent += 
                    '     xmlns:' + namespacePair[0] + '="' + namespacePair[1] + '"\n';
            }
            
            rdfFileInitialContent += '</rdf:RDF>\n';
            
            if(!rdfFile.write(rdfFileInitialContent))
            {
                rdfFile.close();
                return null;
            }
            
            rdfFile.close();
    }
        
        return new ceRDFDataModel(filePathURI, loadObserver);
    }
};


